
#include <iostream>
#include <fstream>
#include <iomanip>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "auxiliary.h"

using namespace std;

// A simple routine to read data from .WAV files
int AudioWav::read(const char* fname, char endian)
{
  char szChunckID[5] = ""; //four bytes to hold chunck id "RIFF" etc
  short format_tag, channels, block_align, bits_per_sample;
  int byte_rate, chunck_size, count(0);

  ifstream in(fname, ios::in | ios::binary);
  if (!in) { 
    cerr << "Unable to open wave file: " << fname << endl; 
    return -1;
  }
  in.read(szChunckID, 4);
  if (strcmp(szChunckID, "RIFF")) { 
    cerr << "Not a RIFF (WAVE) file: found " << szChunckID << endl;
  }
  else {
    in.read((char*)&chunck_size, sizeof(int)); //read in 32bit size value
    in.read(szChunckID, 4); //read in 4 bytes "WAVE"
    if ( strcmp(szChunckID, "WAVE"))
      cerr << "RIFF file but not a wave file: found " << szChunckID << endl;
    else {
      in.read(szChunckID, 4); //read in 4 bytes "fmt ";
      in.read((char*)&chunck_size, sizeof(int));
      in.read((char*)&format_tag, sizeof(short));
      in.read((char*)&channels, sizeof(short));
      in.read((char*)&_fsHz, sizeof(int));
      in.read((char*)&byte_rate, sizeof(int));
      in.read((char*)&block_align, sizeof(short));
      in.read((char*)&bits_per_sample, sizeof(short)); //8 bit or 16 bit file?
      in.read(szChunckID, 4); //read in 4 bytes "data"
      in.read((char*)&chunck_size, sizeof(int)); //sound data in bytes
      int len = chunck_size / (bits_per_sample / 8);
      _data.clear();
      short sample;
      for (count=0; count<len; count++) {
        in.read((char*)&sample, sizeof(short));
        if (in.gcount() == sizeof(short)) {
          if (endian == 'b')
            swap_2bytes ( &sample );
          // normalisation as in Matlab "wavread"
          _data.push_back(sample/32768.0);
        }
        else {
          cerr << "Unable to read " << len << " samples: read " << count+1 << endl;
          break;
        }
      }
    }
  }
  in.close();
  return count;
}

int read_float32(MatrixF & mat, const char* fn, int numChans, int offset, bool byteswap)
{
  ifstream file(fn, ios::in|ios::binary);
  if (!file.is_open()) {
    cerr << "Unable to open wave file: " << fn << endl;
    return -1;
  }
  file.seekg(0, ios::end);
  int fsize = file.tellg();
  int numFrames = (fsize - offset) / 4 / numChans;
  file.seekg(offset, ios::beg);

  VectorF frame(numChans);
  float f;
  mat.clear();
  for (int i=0; i<numFrames; i++) {  
    for (int j=0; j<numChans; j++) {  
      file.read((char*)&f, 4);
      if ( byteswap ) 
        swap_4bytes((float*)&f);
      frame[j] = f;
    }
    mat.push_back(frame);
  }
  file.close();
  return numFrames;
}

// Assuming log10 compressed ratemaps
//   mix    - mixture of speech and noise
//   noise  - estimated noise
void compute_snrmap_log(MatrixF& snrmap, const MatrixF& mix, const MatrixF& noise)
{
  snrmap.clear();
  int numFrames = mix.size();
  int numChans = mix[0].size();

  VectorF frame(numChans);
  for (int i=0; i<numFrames; i++) {
    for (int j=0; j<numChans; j++) {
      float f = pow(10.0f, mix[i][j]) - pow(10.0f, noise[i][j]);
      if (f < 0.0f) f = 1e-100;
      frame[j] = 20.0f * (log10(f) - noise[i][j]);
    }
    snrmap.push_back(frame);
  }
}

void compute_softmask_sigmoid(MatrixF& snrmask, const MatrixF& snrmap, const float snrthr, const float slope)
{
  snrmask.clear();
  int numFrames = snrmap.size();
  int numChans = snrmap[0].size();
  
  VectorF frame(numChans);
  for (int i=0; i<numFrames; i++) {
    for (int j=0; j<numChans; j++) {
      frame[j] = 1.0 / (1.0 + exp(-slope * (snrmap[i][j] - snrthr)));
    }
    snrmask.push_back(frame);
  }
}

void print_vector ( const int* x, const int len )
{
   for ( int i = 0; i < len ; i++ ) {
      if ( i % 16 == 0 ) cout << endl; // << endl << "c" << setiosflags(ios::left) << setw(3) << i + 1 << ": ";
      cout << setiosflags(ios::left) << setw(5) << x[i];
   }
   cout << endl;
}

void print_vector ( const Float* x, const int len )
{
   for ( int i = 0; i < len ; i++ ) {
      if ( i % 16 == 0 ) cout << endl; // << endl << "c" << setiosflags(ios::left) << setw(3) << i + 1 << ": ";
      cout << setiosflags(ios::left) << setw(8) << setiosflags(ios::fixed) << setprecision(2) << x[i];
   }
   cout << endl;
}

// Dealing with the little/big endian swapping
void swap_2bytes(short* x)
{
   char* p = (char*)x;
   char c = *p; *p = *(p+1); *(p+1) = c;
}

void swap_4bytes(float* x)
{
   char* p = (char*)x;
   char c = *p; *p = *(p+3); *(p+3) = c;
   c = *(p+1); *(p+1) = *(p+2); *(p+2) = c;
}


// end

