
#include <iostream>
#include <algorithm>
#include <vector>
#include <utility>
#include <fstream>
#include <math.h>
#include <string.h>
#include "auxiliary.h"
#include "casagrouping.h"

using namespace std;

const int MIN_PRE_GROUPS = 10;           // If too many channels look similar, the frame may well be unvoiced...???
const int opHalfWidth = 30;              // Gabor operator width depends on the sampling rate and 
const int opWidth = opHalfWidth * 2 + 1; // should cover at lease 3 peaks in an ACG channel
const int opHalfSize = 2;                // For 64 channel ACG. Increase it if more channels are employed
const int opSize = opHalfSize * 2 + 1;
const int NCHANS_MIN = 4;


CASAGrouping::CASAGrouping(const AcgModel& a, PitchTracker& t) :
  _acg(a), _tracker(t), _mdmask(NULL), _curframe(0), _globalpeak(0),
  _frmseg(a.nframes(),VectorInt(a.nchans(),0)), _f0cand(a.nframes(),VectorF(NUM_F0CANDS,0.0)),
  _labels(a.nframes(),VectorInt(NUM_SOURCES,0)), _tracks(a.nframes(),VectorF(NUM_SOURCES,0.0)),
  _groups(a.nframes(),VectorInt(a.nchans(),0)), _masks(a.nframes(),VectorF(a.nchans(),0.0)),
  _sacg(a.maxdelay(),0.0), _gacg(a.nchans(),VectorF(a.maxdelay(),0.0)),
  _gcos(a.nchans(),VectorF(opHalfWidth,0.0)), _gsin(a.nchans(),VectorF(opHalfWidth,0.0))
{
  _minLag = (int) ceil ( _acg.sr() / (Float) MAX_F0 ) + 1;
  _maxLag = (int) round ( _acg.sr() / (Float) MIN_F0 ) - 1;
  _maxLag = min ( _maxLag, _acg.maxdelay() );
}

// ++ Spectral integration
// ++ Pitch tracking
// ++ Temporal integration
void CASAGrouping::generate_fragments(const MatrixF* mdmask)
{
  _mdmask = mdmask;
  spectral_integrate();
  _tracker.track(_f0cand, _acg.nframes(), _labels, _tracks);
  form_fragments();
}

void CASAGrouping::write_fragments(const char* outfn) const
{
  ofstream out(outfn, ios::out | ios::binary);
  if (!out) { 
    cerr << "Unable to open file: " << outfn << endl; 
    return;
  }
  // Write fragments as shorts (int16)
  for (size_t i=0; i<_groups.size(); i++) {
    for (size_t j=0; j<_groups[0].size(); j++) {
      short s = static_cast<short>(_groups[i][j]);
      out.write((char*)&s, sizeof(short));
    }
  }
  out.close(); 
}

void CASAGrouping::write_mask(const char* outfn) const
{
  ofstream out(outfn, ios::out | ios::binary);
  if (!out) { 
    cerr << "Unable to open file: " << outfn << endl; 
    return;
  }
  // Write masks as floats (float32)
  for (size_t i=0; i<_masks.size(); i++) {
    for (size_t j=0; j<_masks[0].size(); j++) {
      float f = static_cast<float>(_masks[i][j]);
      out.write((char*)&f, sizeof(float));
    }
  }
  out.close(); 
}

void CASAGrouping::write_frmSegments(const char* outfn) const
{
  ofstream out(outfn, ios::out | ios::binary);
  if (!out) { 
    cerr << "Unable to open file: " << outfn << endl; 
    return;
  }
  // Write Frame Segmentations as shorts (int16)
  for (size_t i=0; i<_frmseg.size(); i++) {
    for (size_t j=0; j<_frmseg[0].size(); j++) {
      short s = static_cast<short>(_frmseg[i][j]);
      out.write((char*)&s, sizeof(short));
    }
  }
  out.close();
}

void CASAGrouping::write_f0Candidates(const char* outfn) const
{
  ofstream out(outfn, ios::out | ios::binary);
  if (!out) { 
    cerr << "Unable to open file: " << outfn << endl; 
    return;
  }
  // Write F0 candidates as floats (float32)
  for (size_t i=0; i<_f0cand.size(); i++) {
    for (size_t j=0; j<_f0cand[0].size(); j++) {
      float f = static_cast<float>(_f0cand[i][j]);
      out.write((char*)&f, sizeof(float));
    }
  }
  out.close(); 
}

void CASAGrouping::write_f0Tracks(const char* outfn) const
{
  ofstream out(outfn, ios::out | ios::binary);
  if (!out) { 
    cerr << "Unable to open file: " << outfn << endl; 
    return;
  }
  // Write F0 Tracks as floats
  for (size_t i=0; i<_tracks.size(); i++) {
    for (size_t j=0; j<_tracks[0].size(); j++) {
     float f = static_cast<float>(_tracks[i][j]);
     out.write((char*)&f, sizeof(float));
    }
  }
  out.close(); 
}

void CASAGrouping::write_f0Labels(const char* outfn) const
{
  ofstream out(outfn, ios::out | ios::binary);
  if (!out) { 
    cerr << "Unable to open file: " << outfn << endl; 
    return;
  }
  // Write labels as shorts (int16)
  for (size_t i=0; i<_labels.size(); i++) {
    for (size_t j=0; j<_labels[0].size(); j++) {
      short s = static_cast<short>(_labels[i][j]);
      out.write((char*)&s, sizeof(short));
    }
  }
  out.close(); 
}

// Form harmonic fragments based on pitch tracks
void CASAGrouping::form_fragments()
{
  static const int f0range = 20; // in Hz
  const int nlags = _acg.maxdelay();
  Float macg, f0shift, ratio; int halfL, halfR, shift;
  for (int t = 0; t < _acg.nframes(); t++) {
    for (int i = 0; i < NUM_SOURCES; i++) {
      if ( _labels[t][i] > 0 ) {
        // Spectral group 1
        if ( _tracks[t][i] == _f0cand[t][0] ) {
          halfL = (int)round ( _acg.sr() / ( _tracks[t][i] + f0range ) );
          halfR = min ( nlags-1, (int)round (_acg.sr()/(_tracks[t][i]-f0range)) );
          for ( int c = 0; c < _acg.nchans(); c++ ) {
            if (_frmseg[t][c] == GROUP1 ) {
              _groups[t][c] = _labels[t][i];
              macg = 0.0; shift = halfL;
              for ( int d = halfL; d <= halfR; d++ ) {
                if ( macg < _acg.get(t, c, d)) {
                  macg = _acg.get(t, c, d);
		  shift = d;
                }
              }
              f0shift = fabs ( _tracks[t][i] - _acg.sr() / (float)shift );
              ratio = 1 / ( 1 + expf( -0.15 * (f0range-f0shift) ) );
              if ( ratio < 0.5) ratio = 0.501;
               _masks[t][c] = ratio;
            }
          }
        }
        // Spectral group 2
        if (_tracks[t][i] == _f0cand[t][1] || _tracks[t][i] == _f0cand[t][2] || _tracks[t][i] == _f0cand[t][3]) {
          if (round(_acg.sr() / _tracks[t][i]) > _maxLag) {
            for (int c = 0; c < _acg.nchans(); c++) {
              if (_frmseg[t][c] == 2) {
                _groups[t][c] = _labels[t][i];
                _masks[t][c] = 0.501;
              }
            }
          }
          else {
            halfL = (int) round(_acg.sr() / ( _tracks[t][i] + f0range));
            halfR = min(nlags-1, (int)round (_acg.sr()/(_tracks[t][i]-f0range)));
            for ( int c = 0; c < _acg.nchans(); c++ ) {
              if (_frmseg[t][c] == GROUP2) {
                _groups[t][c] = _labels[t][i];
                macg = 0.0;
                shift = halfL;
                for (int d = halfL; d <= halfR; d++) {
                  if ( macg < _acg.get(t, c, d)) {
                    macg = _acg.get(t, c, d); shift = d;
                  }
                }
                f0shift = fabs ( _tracks[t][i] - _acg.sr() / (float)shift );
                ratio = 1 / ( 1 + expf( -0.15 * (f0range-f0shift) ) );
                if ( ratio < 0.5)
                  ratio = 0.501;
                _masks[t][c] = ratio;
              }
            }
          }
        }
      }
    }
  }
}

/****************************************************************
 * Spectral integration:
 *   group channels in each frame into spectral groups and
 *   compute F0 candidates for each spectral group
 ****************************************************************/
void CASAGrouping::spectral_integrate ()
{
  for ( _curframe = 0; _curframe < _acg.nframes(); _curframe++ ) {
    // Check if the frame is silence or unvoiced
    if ( is_silent () || is_unvoiced () )
      continue;

    // Normalise ACG by lag zero energy and copy it to _gacg
    normalise_current_acg();

    // Group channels by looking for the tree structure in the 2D ACG
    group_channels();
  }
}

// Group pre-grouped channels by looking for the tree structure in the 2D ACG
void CASAGrouping::group_channels()
{
  updateGabor();
   
  VectorInt spineLags(_acg.nchans());
  extract_dendrites(spineLags);

  //print_vector(&spineLags[0], spineLags.size());

  int nTrees = 2;
  VectorInt spines(4, 0); // The last 2 spaces are for double spines.
  bool useSummary = locate_dendrites(spineLags, spines, nTrees);

  VectorInt spineGrps(_acg.nchans(), NOGROUP);
  // Adjust spines (and spines double) according to the actual peak position in the summary
  // of grouped channels
  adjust_spines(spineLags, spineGrps, useSummary, spines, nTrees);

  if (useSummary && _globalpeak > _maxLag)
    return;

  // Assign channels to each spine
  assign_channels(spines, nTrees, spineGrps);

  // Update segments and F0 candidates for each spectral group
  update_FrmSeg_F0Cand(spineGrps);
}

// Compute F0 candidates for each spectral group
void CASAGrouping::update_FrmSeg_F0Cand(const VectorInt& spineGrps)
{
  int cnts[2] = {0, 0}, c, peaksIdx[MAX_NPEAKS], npeaks, i, peak;
  const int SOURCE_MIN_CHANS = 1, deltaF0 = 3; Float m;
  for ( c = 0; c < _acg.nchans(); c++ ) {
    if ( (i=spineGrps[c]) == NOGROUP )
      continue;
    cnts[i-1]++;
    _frmseg[_curframe][c] = i;
  }

  // Group 1
  if ( cnts[0] > SOURCE_MIN_CHANS ) {
    sum_acg(GROUP1);
    npeaks = pick_peaks(&_sacg[0], _acg.maxdelay(), 15, _minLag, peaksIdx, MAX_NPEAKS);
    if ( npeaks > 0 ) {
      // Find the largest peak
      peak = peaksIdx[0]; m = _sacg[peak]; 
      for ( i = 1; i < npeaks; i++ ) {
        if ( _sacg[peaksIdx[i]] > m ) {
          peak = peaksIdx[i]; m = _sacg[peak];
        }
      }
      _f0cand[_curframe][0] = _acg.sr() / (float) peak;
    }
    else cnts[0] = 0;
  }
  else cnts[0] = 0;

  // Group 2
  if ( cnts[1] > SOURCE_MIN_CHANS ) {
    sum_acg(GROUP2);
    npeaks = pick_peaks(&_sacg[0], _acg.maxdelay(), 15, _minLag, peaksIdx, MAX_NPEAKS);
    if ( npeaks > 0 ) {
      vector<pair<Float, int> > vec;
      for ( i = 0; i < npeaks; i++ )
	vec.push_back(pair<Float, int>(_sacg[peaksIdx[i]], i));
      sort(vec.begin(), vec.end());
      vector<pair<Float, int> >::const_iterator itr = vec.begin() + npeaks - 1;
      _f0cand[_curframe][1] = _acg.sr() / (Float) peaksIdx[(*itr).second];
      if ( npeaks > 1 && (*(itr-1)).first > (*itr).first * 0.9 ) {
        _f0cand[_curframe][2] = _acg.sr() / (Float) peaksIdx[(*(itr-1)).second];
        if ( fabs ( _f0cand[_curframe][1] - _f0cand[_curframe][2] * 2 ) < deltaF0 )
          _f0cand[_curframe][2] = 0.0;
        if ( fabs ( _f0cand[_curframe][1] - _f0cand[_curframe][2] ) < deltaF0 )
          _f0cand[_curframe][2] = 0.0;
      }
      if ( npeaks > 2 && (*(itr-2)).first > (*itr).first * 0.9 ) {
        _f0cand[_curframe][3] = _acg.sr() / (float) peaksIdx[(*(itr-2)).second];
        if ( fabs ( _f0cand[_curframe][1] - _f0cand[_curframe][3] * 2 ) < deltaF0 )
          _f0cand[_curframe][3] = 0.0;
        if ( fabs ( _f0cand[_curframe][1] - _f0cand[_curframe][3] ) < deltaF0 )
          _f0cand[_curframe][3] = 0.0;
      }
    }
    else { // Use the largest peak
      find_max(&_sacg[0], _minLag, _maxLag, peak);
      _f0cand[_curframe][1] = _acg.sr() / (float) peak;
    }
  }
  else {
    cnts[1] = 0;
  }

  bool oneGroup = false;
  if ( _f0cand[_curframe][0] > 0 && _f0cand[_curframe][1] > 0 ) {
    if ( fabs ( _f0cand[_curframe][0] - _f0cand[_curframe][1] ) < deltaF0 ) {
      _f0cand[_curframe][1] = _f0cand[_curframe][2] = _f0cand[_curframe][3] = 0.0;
      oneGroup = true;
    }
  }
  if ( _f0cand[_curframe][0] > 0 && _f0cand[_curframe][2] > 0 ) {
    if ( fabs ( _f0cand[_curframe][0] - _f0cand[_curframe][2] ) < deltaF0 )
      _f0cand[_curframe][2] = 0.0;
    else if ( fabs ( _f0cand[_curframe][0] - _f0cand[_curframe][2] * 2 ) < deltaF0 )
      _f0cand[_curframe][2] = 0.0;
    else if ( fabs ( _f0cand[_curframe][0] * 2 - _f0cand[_curframe][2] ) < deltaF0 )
      _f0cand[_curframe][2] = 0.0;
  }
  if ( _f0cand[_curframe][0] > 0 && _f0cand[_curframe][3] > 0 ) {
    if ( fabs ( _f0cand[_curframe][0] - _f0cand[_curframe][3] ) < deltaF0 )
      _f0cand[_curframe][3] = 0.0;
    else if ( fabs ( _f0cand[_curframe][0] - _f0cand[_curframe][3] * 2 ) < deltaF0 )
      _f0cand[_curframe][3] = 0.0;
    else if ( fabs ( _f0cand[_curframe][0] * 2 - _f0cand[_curframe][3] ) < deltaF0 )
      _f0cand[_curframe][3] = 0.0;
  }

  // _frmseg has been set to zeros already in init_spectral()
  for ( c = 0; c < _acg.nchans(); c++ ) {
    if ( (i = spineGrps[c]) == NOGROUP ) {
      _frmseg[_curframe][c] = 0;
      continue;
    }
    if (oneGroup)
      i = GROUP1;
    if (cnts[i-1] > 0)
      _frmseg[_curframe][c] = i;
  }
}

// Assign channels to each spectral groups
void CASAGrouping::assign_channels(const VectorInt& spines, const int nTrees, VectorInt& spineGrps)
{
  for (size_t i=0; i<spineGrps.size(); i++)
    spineGrps[i] = NOGROUP;

  const float CF1 = 740, CF2 = 3650;
  Float threshold = _acg.sr() * 0.05 / (Float) spines[0];
  int r1a, r1b, r2a(0), r2b(0);
  rangeInHz(spines[0], threshold * 2, r1a, r1b);
  if ( nTrees > 1 ) {
    threshold = min(threshold, _acg.sr() * 0.05 / (Float)spines[1]);
    rangeInHz(spines[1], _acg.sr() * 0.1 / (Float)spines[1], r2a, r2b);
  }
  Float m1, m2, shift1, shift2; int midx, sp1, sp2 = 0;
  for ( int c = 0; c < _acg.nchans(); c++) {
    // The first tree
    m1 = find_max(&_gacg[c][0], r1a, r1b, midx);
    if ( _gacg[c][0] == 0 || m1 / _gacg[c][0] < 0.1 ) continue;
    sp1 = spines[0];

    if ( _acg.cf(c) > CF1 ) { // Hz
      if ( _gacg[c][sp1-1] > _gacg[c][sp1] ) sp1--;
      else if ( _gacg[c][sp1+1] > _gacg[c][sp1] ) sp1++;
      if ( _gacg[c][sp1] / m1 > 0.95 && _gacg[c][sp1] / _gacg[c][0] > 0.5 )
        midx = sp1;
    }
    shift1 = calc_distHz(midx, sp1);

    // Energy based threshold
    if ( m1 > 0 && _gacg[c][sp1] / m1 > 0.95 && _gacg[c][sp1] / _gacg[c][0] > 0.5 ) {
      if ( shift1 > threshold - 1 ) shift1 = threshold - 1;
    }
    else { shift1 = threshold + 1; }

    // The second tree
    shift2 = threshold + 1; m2 = 0;
    if ( nTrees > 1 ) {
      m2 = find_max(&_gacg[c][0], r2a, r2b, midx);
      sp2 = spines[1];
      if ( _acg.cf(c) > CF1 ) { // Hz
        if ( _gacg[c][sp2-1] > _gacg[c][sp2] ) sp2--;
        else if ( _gacg[c][sp2+1] > _gacg[c][sp2] ) sp2++;
        if ( _gacg[c][sp2] / m2 > 0.95 && _gacg[c][sp2] / _gacg[c][0] > 0.5 )
          midx = sp2;
      }
      shift2 = calc_distHz ( midx, sp2 );
      if ( m2 > 0 && _gacg[c][sp2] / m2 > 0.95 && _gacg[c][sp2] / _gacg[c][0] > 0.5 ) {
        if ( shift2 > threshold - 1 ) shift2 = threshold - 1;
        if ( _gacg[c][sp2] > _gacg[c][sp1] && shift2 > shift1 ) shift2 = shift1;
      }
    }
     
    // Assign channels...
    if ( nTrees > 1 && _acg.cf(c) > CF2 ) {
      // For high freq channels we use energy. this really should be replaced by AM envelope
      if ( _gacg[c][sp1] > _gacg[c][sp2] ) spineGrps[c] = GROUP1;
      else spineGrps[c] = GROUP2;
    }
    else if ( shift1 < threshold && ( shift1 < shift2 || m1 > m2 ) ) {
      // Also check the energy at the lag of "double spine1" to avoid pitch doubling error
      if ( _gacg[c][sp1] > _gacg[c][spines[2]] ) spineGrps[c] = GROUP1;
      else if ( nTrees > 1 ) spineGrps[c] = GROUP2;
    }
    else if ( shift2 < threshold && ( shift2 < shift1 || m1 < m2 ) ) {
      spineGrps[c] = GROUP2;
    }
    else if ( nTrees > 1 && _gacg[c][0] > MIN_ENERGY_THRESHOLD ) {
      // Then we have to assign channel g to some group
      if ( _gacg[c][sp1] > _gacg[c][sp2] ) spineGrps[c] = GROUP1;
      else spineGrps[c] = GROUP2;
    }
    // If only 1 tree is found, put the rest channels together into group2
    if ( nTrees == 1 && spineGrps[c] == NOGROUP )
      spineGrps[c] = GROUP2;
  }
}

// Adjust spines according to the actual peak position in the summary
// of grouped channels
void CASAGrouping::adjust_spines(const VectorInt& spineLags, VectorInt& spineGrps, bool& useSummary, VectorInt& spines, int& nTrees)
{
  const int sp1 = 0, sp2 = 1, sp1d = 2, sp2d = 3;
  // Assign spectral groups
  if ( !useSummary ) {
    int r1, r2, r11, r22, ngrp1 = 0, ngrp2 = 0;
    rangeInHz ( spines[sp1], 1, r1, r2 );
    rangeInHz ( spines[sp2], 1, r11, r22 );
    for ( int c = 0; c < _acg.nchans(); c++ ) {
      if ( spineLags[c] >= r1 && spineLags[c] <= r2 ) {
        spineGrps[c] = GROUP1; ngrp1++;
      }
      else if ( nTrees > 1 && spineLags[c] >= r11 && spineLags[c] <= r22 ) {
        spineGrps[c] = GROUP2; ngrp2++;
      }
    }
    // If the strongest dendrite has less then NCHANS_MIN channels, use the global peak instead
    if ( ngrp1 < 3 ) { spines[sp1] = _globalpeak; useSummary = true; nTrees = 1; }
    else if ( ngrp2 < 3 ) { nTrees = 1; }
  }

  // Adjust spines according to the actual peak position in the summary
  const int half = 5;
  if ( useSummary ) {
    spines[sp1d] = spines[sp1] * 2;
    if ( spines[sp1d] + half < _maxLag )
      find_max(&_sacg[0], spines[sp1d] - half, spines[sp1d] + half, spines[sp1d] );
    else
      spines[sp1d] = min(spines[sp1d], _maxLag-1);
  }
  else {
    sum_acg(spineGrps, GROUP1);
    int r1, r2;
    rangeInHz(spines[sp1], half, r1, r2);
    find_max(&_sacg[0], r1, r2, spines[sp1] );

    spines[sp1d] = spines[sp1] * 2;
    if (spines[sp1d] + half < _maxLag)
      find_max(&_sacg[0], spines[sp1d] - half, spines[sp1d] + half, spines[sp1d]);
    else
      spines[sp1d] = min ( spines[sp1d], _maxLag - 1 );
      
    if ( nTrees > 1 ) {
      if ( spines[sp2] + 2 < _maxLag ) {
        sum_acg(spineGrps, GROUP2);
        rangeInHz ( spines[sp2], half, r1, r2 );
        find_max(&_sacg[0], r1, r2, spines[sp2]);
        spines[sp2d] = spines[sp2] * 2;
        if ( spines[sp2d] + half < _maxLag )
          find_max(&_sacg[0], spines[sp2d] - half, spines[sp2d] + half, spines[sp2d]);
        else
          spines[sp2d] = min ( spines[sp2d], _maxLag - 1 );
      }
      else { nTrees = 1; }
         
      // Check if the two spines are too close and one is a double to the other
      const int DIST_MIN = 3; // in Hz
      if ( calc_distHz(spines[sp1], spines[sp2]) <= DIST_MIN ) {
        nTrees = 1;
      }
      else if ( calc_distHz(spines[sp1], spines[sp2] * 2) < DIST_MIN - 1 ) {
        nTrees = 1; spines[sp1] = spines[sp2]; spines[sp1d] = spines[sp2d];
      }
      else if ( calc_distHz(spines[sp2], spines[sp1] * 2) < DIST_MIN - 1 ) {
        nTrees = 1;
      }
    }
  }
}

// Extract tree structures in ACG (2D convolution with Gabors)
void CASAGrouping::extract_dendrites(VectorInt& spineLags)
{
  // Locate spines in the channels at the edges
  for (int c = 0; c < opHalfSize; c++ ) {
    int midx = _minLag;
    Float macg = 0.0;
    for (int i = _minLag; i < _maxLag; i++ ) {
      Float d = _gacg[c][i];
      if ( d > macg ) { macg = d; midx = i; }
    }
    spineLags[c] = midx;
    midx = _minLag;
    macg = 0.0;
    for (int i = _minLag; i < _maxLag; i++ ) {
      Float d = _gacg[_acg.nchans()-1-c][i];
      if ( d > macg ) { macg = d; midx = i; }
    }
    spineLags[_acg.nchans()-1-c] = midx;
  }
  // Locate spines in the rest channels by performing 2D convolutions with a Gabor kernel
  for (int c = opHalfSize; c < _acg.nchans() - opHalfSize; c++)
    spineLags[c] = gabor_conv2(c);
}

// Compute C = conv2(A, B) and keep only the centre row of A part in C.
// Only the CONV2 points between _minLag and _maxLag are not computed.
// It emploits the symmetry of Gabor functions (sin with a different sign) to
// save computations.
// Returns the lag corresponding to the location of a potential tree structure
// See myconv2_sym.m for a Matlab implementation
int CASAGrouping::gabor_conv2 ( const int centre )
{
   const int nActLags = _maxLag - _minLag;
   int i, j, cj, n;
   VectorF gxcos(nActLags, 0.0), gxsin(nActLags, 0.0);
   const int n1 = centre+opHalfSize, n2 = centre-opHalfSize;
   Float t1, t2;
   for ( i = 0; i < opSize; i++ ) {
      for ( j = 0; j < opHalfWidth; j++ ) {
         for ( cj = _minLag; cj < _maxLag; cj++ ) {
            n = cj - j + opHalfWidth;
            if ( n >= 0 && n < _acg.maxdelay() ) {
	      t1 = t2 = _gacg[n1-i][n];
	    }
            else {
	      t1 = t2 = 0.0; 
	    }
            n = cj + j - opHalfWidth;
            if ( n >= 0 && n < _acg.maxdelay() ) {
	      t1 += _gacg[n1-i][n];
	      t2 -= _gacg[n1-i][n]; 
	    }
            gxcos[cj-_minLag] += t1 * _gcos[n2+i][j];
            gxsin[cj-_minLag] += t2 * _gsin[n2+i][j];
         }
      }
      // Add the case when j == opHalfWidth: _gcos(opHalfWidth) == 1, _gsin(opHalfWidth) == 0;
      for ( cj = _minLag; cj < _maxLag; cj++ )
         gxcos[cj-_minLag] += _gacg[n1-i][cj];
   }
   int midx = 0; Float d, macg = 0.0;
   for ( i = 0; i < nActLags - 1; i++ ) {
      d = gxcos[i] * gxcos[i] + gxsin[i] * gxsin[i];
      if ( d > macg ) { macg = d; midx = i; }
   }
   return (midx + _minLag);
}

// Locate the tree structures by computing a histogram
// Return true if no trees are found and the summary ACG is used instead
bool CASAGrouping::locate_dendrites(const VectorInt& spineLags, VectorInt& spines, int& nTrees)
{
  // Compute a histogram of spine indexes
  const int lagBegin = _minLag + 3, lagEnd = _maxLag - 3, nlags = lagEnd - lagBegin + 1;
  VectorF lagHist(nlags, 0.0);
  for (int c = 0; c < _acg.nchans(); c++) {
    int lag = spineLags[c] - lagBegin;
    if (lag >= 0 && lag < nlags) {
      if (NULL==_mdmask)
        lagHist[lag]++;
      else
	lagHist[lag] += _mdmask->at(_curframe)[c];
    }
  }
 
  nTrees = 1; spines[0] = -1, spines[0] = -1;
  bool useSummary = false;

  // Try to find the largest tree
  int r1, r2, j;
  float mhist(0.0), nchan1(0.0), nchan2(0.0);
  for (int lag = 2; lag < nlags; lag++ ) {
    rangeInHz(lag+lagBegin, 1, r1, r2);
    r1 -= lagBegin; r2 -= lagBegin; mhist = 0.0;
    for (j = r1; j <= r2; j++ ) {
      if ( j >= 0 && j < nlags ) mhist += lagHist[j];
    }
    if ( mhist > nchan1 ) { nchan1 = mhist; spines[0] = lag; }
  }   
  if ( spines[0] < 0 ) { spines[0] = _globalpeak; useSummary = true; }
  else {
    spines[0] += lagBegin;
    // Try to find the second largest tree
    // Set spine1 to zero first
    rangeInHz ( spines[0], 1, r1, r2 );
    for (int lag = r1; lag <= r2; lag++ )
      lagHist[lag-lagBegin] = 0;
    for (int lag = 0; lag < nlags; lag++ ) {
      rangeInHz(lag+lagBegin, 1, r1, r2);
      r1 -= lagBegin; r2 -= lagBegin; mhist = 0;
      for (j = r1; j <= r2; j++) {
        if (j >= 0 && j < nlags) mhist += lagHist[j];
      }
      if ( mhist > nchan2 ) {
        // Check if the second spine is too close or a doubling of the first spine
        if (calc_distHz(spines[0], lag+lagBegin) > 3 && calc_distHz(spines[0], (lag+lagBegin)/2) > 3 ) {
          nchan2 = mhist; spines[1] = lag;
        }
      }
    }
  }
  if ( spines[1] >= 0 ) { nTrees = 2; spines[1] += lagBegin; }
      
  // If the strongest dendrite has less then NCHANS_MIN channels, use the global peak instead
  //if ( nchan1 < NCHANS_MIN ) { spines[0] = _globalpeak; useSummary = true; }
  //else if ( nchan2 < NCHANS_MIN ) { nTrees = 1; }

  return useSummary;
}

// Update Gabor cosine & sine functions
void CASAGrouping::updateGabor()
{
   int idx1peak, idx2peak, idx1valley;
   Float sigma, gabor, chnlPeriod;
   for (int c = 0; c < _acg.nchans(); c++) {
      // Find channel periods
      idx1peak = idx2peak = idx1valley = 0;
      for (int i = 2; i < _acg.maxdelay() - 10; i++ ) {
         if ( idx1peak == 0 && _gacg[c][i] > _gacg[c][i-1] && _gacg[c][i] > _gacg[c][i+1] ) {
            if ( _gacg[c][i] > _gacg[c][0]*0.3 )
               idx1peak = i;
            else if ( idx2peak == 0 )
               idx2peak = i;
         }
         if ( idx1valley == 0 && _gacg[c][i] < _gacg[c][i-1] && _gacg[c][i] < _gacg[c][i+1] )
            idx1valley = i;
         if ( idx1peak > 0 && idx1valley > 0 && idx2peak > 0 )
            break;
      }
      if ( idx1peak == 0 ) idx1peak = idx2peak;
      chnlPeriod = ( idx1valley * 2.0 + idx1peak ) / 2.0;
      if ( chnlPeriod < 1.0 ) chnlPeriod = 1.0;

      // Calculate Gabor functions based on channel periods
      sigma = chnlPeriod * 0.4;
      for (int i = 0, x = -opHalfWidth; i < opHalfWidth; i++, x++ ) {
         gabor = exp(-0.5*x*x/(sigma*sigma));
         _gcos[c][i] = gabor * cos(2 * M_PI * x / chnlPeriod);
         _gsin[c][i] = gabor * sin(2 * M_PI * x / chnlPeriod);
      }
   }
   //print_vector(&_gcos[10][0], _gcos[10].size());
}

// Check if the current frame is silent
bool CASAGrouping::is_silent()
{
   int maxChan = 0;
   Float curEnergy, maxEnergy = 0.0;
   for ( int chan = 0; chan < _acg.nchans(); chan++ ) {
      curEnergy = _acg.get(_curframe, chan, 0);
      if ( curEnergy > maxEnergy ) {
         maxEnergy = curEnergy;
         maxChan = chan+1;
      }
   }
   if ( maxEnergy < MIN_ENERGY_THRESHOLD || maxChan < LOW_FREQ_CHANNEL )
      return true;
   else
      return false;
}

// A simple technique to check if the current frame is unvoiced (low harmonisity)
// The current frame is unvoiced if the ratio of the largest peak to lag zero 
// energy is less than 0.5
bool CASAGrouping::is_unvoiced ()
{
   // Compute summary ACG
   sum_acg();
   _globalpeak = -1;
   bool unvoiced = false;
   // Look for peaks in the summary ACG (sacg)
   int peaksidx[MAX_NPEAKS];
   int npeaks = pick_peaks(&_sacg[0], _acg.maxdelay(), 20, _minLag, peaksidx, MAX_NPEAKS);
   if ( 0 == npeaks )
      unvoiced = true;
   else {
      Float peakenergy = 0.0;
      for ( int i = 0; i < npeaks; i++ ) {
         if ( peakenergy < _sacg[peaksidx[i]] ) {
            peakenergy = _sacg[peaksidx[i]];
            _globalpeak = peaksidx[i];
         }
      }
      if ( _globalpeak < _minLag + 3 || peakenergy / _sacg[0] < 0.5 )
         unvoiced = true;
   }
   return unvoiced;
}

// Normalise the ACG at the current frame, store it in _gacg
void CASAGrouping::normalise_current_acg()
{
  // Copy original ACG to _gacg and normliase it by lag zero energy
  for (int chan = 0; chan < _acg.nchans(); chan++) {
    _gacg[chan][0] = 1.0;
    Float norm = _acg.get(_curframe, chan, 0);
    if (norm <= 0) norm = 1.0;
    for (int lag = 1; lag < _acg.maxdelay(); lag++) {
      _gacg[chan][lag] = _acg.get(_curframe, chan, lag) / norm;
      if ( _gacg[chan][lag] > 1.0 )
        _gacg[chan][lag] = 1.0;
    }
  }
  //print_vector(&_gacg[0][0], _acg.maxdelay());
}

// Compute summary ACG and save the result in the buffer _sacg
void CASAGrouping::sum_acg()
{
  for (int lag=0; lag < _acg.maxdelay(); lag++)
    _sacg[lag] = 0.0;

  for (int chan = 0; chan < _acg.nchans(); chan++) {
    Float weight = 1.0;
    if (NULL != _mdmask) weight = _mdmask->at(_curframe)[chan];
    for (int lag = 0; lag < _acg.maxdelay(); lag++ )
      _sacg[lag] += weight * _acg.get(_curframe, chan, lag);
  }
}

// Sum ACG channels assigned to group grp
void CASAGrouping::sum_acg(SPINE_GROUPS grp)
{
  for (int lag=0; lag < _acg.maxdelay(); lag++)
    _sacg[lag] = 0.0;

  for (int chan = 0; chan < _acg.nchans(); chan++ ) {
    if ( _frmseg[_curframe][chan] == grp ) {
      Float weight = 1.0;
      if (NULL!=_mdmask) weight = _mdmask->at(_curframe)[chan];
      for (int lag = 0; lag < _acg.maxdelay(); lag++ )
        _sacg[lag] += weight * _acg.get(_curframe, chan, lag);
    }
  }
}

// Sum ACG channels (in _gacg) assigned to group grp according to spineGrps
void CASAGrouping::sum_acg(const VectorInt& spineGrps, SPINE_GROUPS grp)
{
  for (int lag = 0; lag < _acg.maxdelay(); lag++ )
    _sacg[lag] = 0.0;

  for (int chan = 0; chan < _acg.nchans(); chan++ ) {
    if (spineGrps[chan] == grp) {
      for (int lag = 0; lag < _acg.maxdelay(); lag++)
        _sacg[lag] += _gacg[chan][lag];
    }
  }
}

void CASAGrouping::rangeInHz(const int lag, const Float half, int& r1, int& r2)
{
   r1 = (int) floor ( _acg.sr() * lag / (Float) ( _acg.sr() + half * lag ) );
   r2 = (int) ceil ( _acg.sr() * lag / (Float) ( _acg.sr() - half * lag ) );
   if ( r2 >= _maxLag ) r2 = _maxLag - 1;
}

Float CASAGrouping::calc_distHz ( const int lag1, const int lag2 )
{
   return fabs ( _acg.sr() / (Float)lag1 - _acg.sr() / (Float)lag2 );
}

//----------------------------------------------- Static functions

// Return the maximum number in x between [n1, n2], the corresponding index is saved in "mi"
Float CASAGrouping::find_max(const Float* x, const int n1, const int n2, int& mi)
{
   Float m = 0.0; mi = n1;
   for ( int i = n1; i <= n2; i++ ) {
      if ( x[i] > m ) { m = x[i]; mi = i; }
   }
   return m;
}

// Return the maximum number in x between [n1, n2]
Float CASAGrouping::find_max(const Float* x, const int n1, const int n2)
{
   Float m = 0.0; int mi = n1;
   for ( int i = n1; i <= n2; i++ ) {
      if ( x[i] > m ) { m = x[i]; mi = i; }
   }
   return m;
}

// Return the actual number of peaks picked in px
int CASAGrouping::pick_peaks(const Float* px, const int nx, const int nwin, const int offset, int* peakidx, const int maxnpeaks)
{
   int halfwin = (int) ceil ( nwin * 0.5 );
   int shift = (int) floor ( nwin / 3.0 );

   if ( px[0] <= 0 )
      return 0;
      
   int i, j, curnpeaks = 0;
   Float localmax;
   for ( i = offset; i < nx-halfwin-1; i++ ) {
      localmax = 0;
      for ( j = i-halfwin; j <= i+halfwin; j++ ) {
         if ( localmax < px[j] ) localmax = px[j];
      }
      if ( px[i] == localmax ) {
         if ( px[i] > 0.4 * px[0] ) {
            peakidx[curnpeaks] = i;
            curnpeaks++;
            if ( curnpeaks >= maxnpeaks )
               break;
            i += shift;
         }
      }
   }
   return curnpeaks;
}

// end
