
#ifndef __CASAGROUPING_H
#define __CASAGROUPING_H

#include "acgmodel.h"
#include "pitchtracker.h"
#include "auxiliary.h"

const Float MIN_F0           = 71.4; // in Hz (FS=25000, FS/MIN_F0 ~ 350)
const Float MAX_F0           = 400.0;// in Hz
const int HIGH_FREQ_CHANNEL   = 16;   // unreolved channel threshold, based on NUM_CHANNELS and low/high cfs
const int LOW_FREQ_CHANNEL    = 2;    // used to detect CF of Channel 3 is 81.6 Hz
const int MAX_NPEAKS          = 32;   // maximum MAX_NPEAKS peaks in each summary ACG

// Used to detect silence. This is an arbitrary threshold and may change based on waveform energy.
const Float MIN_ENERGY_THRESHOLD = 0.00001;

struct Global {
  const static Float MDMASK_THRESHOLD = 0.5;
};

enum SPINE_GROUPS { NOGROUP = 0, GROUP1, GROUP2 };

class CASAGrouping
{
private:
  const AcgModel&  _acg;
  PitchTracker&    _tracker;
  const MatrixF*   _mdmask;
  int              _curframe;
  int              _globalpeak; // largest peak in the summary ACG

  MatrixInt        _frmseg;     // Segmentation labels for each frame, ACG_NCHANS x nframes, possible values: 0, 1, 2
  MatrixF          _f0cand;     // F0 candidates for each frame, NUM_F0CANDS x nframes
  MatrixInt        _labels;     // Pitch track labels, 2 x nframes
  MatrixF          _tracks;     // Pitch track values, 2 x nframes
  MatrixInt        _groups;     // fragments, ACG_NCHANS x nframes
  MatrixF          _masks;      // soft masks, ACG_NCHANS x nframes

  // temporary buffer
  VectorF          _sacg;       // for summary ACG
  MatrixF          _gacg;       // for normalised ACG at each frame
  MatrixF          _gcos;       // for Gabor cos
  MatrixF          _gsin;       // for Gabor sin

  int              _minLag;
  int              _maxLag;

public:
  CASAGrouping(const AcgModel& a, PitchTracker& t);

   void generate_fragments(const MatrixF* mdmask = NULL);

   void write_fragments(const char* outfn) const;

   void write_mask(const char* outfn) const;
      
   void write_frmSegments(const char* outfn) const;

   void write_f0Candidates(const char* outfn) const;
   
   void write_f0Tracks(const char* outfn) const;
   
   void write_f0Labels(const char* outfn) const;

private:
   void init();

   void form_fragments();

   void spectral_integrate ();

   void group_channels();

   void update_FrmSeg_F0Cand(const VectorInt& spineGrps);

   void extract_dendrites(VectorInt& spineLags);

   bool locate_dendrites(const VectorInt& spineLags, VectorInt& spines, int& nTrees);

   void adjust_spines(const VectorInt& spineLags, VectorInt& spineGrps, bool& useSummary, VectorInt& spines, int& nTrees);

   void assign_channels(const VectorInt& spines, const int nTrees, VectorInt& spineGrps );

   void updateGabor();

   int gabor_conv2(const int centre);

   bool is_unvoiced ();

   bool is_silent ();

   void normalise_current_acg();

   void sum_acg();

   void sum_acg(SPINE_GROUPS grp);

   void sum_acg(const VectorInt& spineGrps, SPINE_GROUPS grp);

   void rangeInHz ( const int lag, const Float half, int& begin, int& end );

   Float calc_distHz ( const int lag1, const int lag2 );

   static int pick_peaks ( const Float* px, const int nx, const int nwin, const int offset, int* peakidx, const int maxnpeaks );

   // Return the maximum number in x between [n1, n2], the corresponding index is saved in "mi"
   static Float find_max ( const Float* x, const int n1, const int n2, int& mi );

   // Return the maximum number in x between [n1, n2]
   static Float find_max ( const Float* x, const int n1, const int n2 );
};


#endif // __CASAGROUPING_H
