
#include <iostream>
#include <algorithm>
#include <string.h>
#include "ruletracker.h"
#include "auxiliary.h"

using namespace std;

// labels and tracks should have at least 2 x nframes spaces
void RuleTracker::track(const MatrixF& f0cands, const int nframes, MatrixInt& labels, MatrixF& tracks)
{
  MatrixF f0s = f0cands;
  remove_closeCands(f0s, nframes);

  // Sort F0 candidates of each frame in ascending order
  for (int t = 0; t < nframes; t++) 
    sort(f0s[t].begin(), f0s[t].end());
  static const int F0CHANGE = 8, lowF0 = 70, labInc = 5;
  int lab1 = 1, lab2 = 1, bmTo, bmFrom, bm;
  float dist1, dist2, dist3, d;
  for (int t = 0; t < nframes - 1; t++) {
    VectorF nextF0s = f0s[t+1];
    for (int n = 0; n < NUM_F0CANDS; n++) {
      if ( f0s[t][n] < lowF0 )
	continue;
      // Step 1: Try to find a candidate match to f0(n,t)
      dist1 = F0CHANGE + 1; bmTo = 0;
      for (int m = 0; m < NUM_F0CANDS; m++) {
        if ( nextF0s[m] < lowF0 )
	  continue;
        d = calc_dist(f0s[t][n], nextF0s[m]);
        if ( d < dist1 ) { dist1 = d; bmTo = m; } // Best match to m
      }
      // Step 1: f0s(bmTo,t+1) is a candidate match to f0s(n,t)
            
      // Step 2: Check if f0(bmTo,t+1) is a definitive match by
      //         looking at the its distance to other f0s at time t
      if ( dist1 < F0CHANGE ) {
        dist2 = F0CHANGE + 1; bmFrom = 0;
        for (int m = n + 1; m < NUM_F0CANDS; m++) {
          if ( f0s[t][m] < lowF0 )
	    continue;
          d = calc_dist(f0s[t][m], nextF0s[bmTo]);
          if ( d < dist2 ) {
	    dist2 = d;
	    bmFrom = m;
	  }
        }
        if ( dist1 <= dist2 ) {
          // Step 2: f0s(bmTo,t+1) is a definitive match to f0s(n,t)
          if (tracks[t][0] == f0s[t][n]) {
            tracks[t+1][0] = nextF0s[bmTo];
            labels[t+1][0] = labels[t][0];
          }
          else if (tracks[t][1] == f0s[t][n]) {
            tracks[t+1][1] = nextF0s[bmTo];
            labels[t+1][1] = labels[t][1];
          }
          else if ( tracks[t][0] == 0 ) {
            // A new track is born
            tracks[t][0] = f0s[t][n];
            tracks[t+1][0] = nextF0s[bmTo];
            lab1 += labInc;
            labels[t+1][0] = labels[t][0] = lab1;
          }
          else if ( tracks[t][1] == 0 ) {
            // A new track is born
            tracks[t][1] = f0s[t][n];
            tracks[t+1][1] = nextF0s[bmTo];
            lab2 += labInc;
            labels[t+1][1] = labels[t][1] = lab2;
          }
          // Eliminate nextF0s[bmTo] from further consideration
          nextF0s[bmTo] = 0.0;
        }
        else {
          // Step 3: check the f0s below f0s(bmTo,t+1) at t+1
          dist3 = F0CHANGE + 1; bm = 0;
          for (int m = 0; m < bmTo - 1; m++) {
            if ( nextF0s[m] < lowF0 ) continue;
            d = calc_dist(f0s[t][n], nextF0s[m]);
            if ( d < dist3 ) { dist3 = d; bm = m; }
          }
          if ( dist3 < F0CHANGE ) {
            // Here we assume f0(bm,t+1) is then a definitive match to f0(n,t)
            if ( tracks[t][0] == f0s[t][n] ) {
              tracks[t+1][0] = nextF0s[bm];
              labels[t+1][0] = labels[t][0];
            }
            else if ( tracks[t][1] == f0s[t][n] ) {
              tracks[t+1][1] = nextF0s[bm];
              labels[t+1][1] = labels[t][1];
            }
            else if ( tracks[t][0] == 0 ) {
              // A new track is born
              tracks[t][0] = f0s[t][n];
              tracks[t+1][0] = nextF0s[bm];
              lab1 += labInc;
              labels[t+1][0] = labels[t][0] = lab1;
            }
            else if ( tracks[t][1] == 0 ) {
              // A new track is born
              tracks[t][1] = f0s[t][n];
              tracks[t+1][1] = nextF0s[bm];
              lab2 += labInc;
              labels[t+1][1] = labels[t][1] = lab2;
            }
            // Eliminate f0s_next(bm) from further consideration
            nextF0s[bm] = 0.0;
          }
        } 
      }
    }
  }
   
  // Get rid of short tracks (2-frame long) if they cannot be combined with
  // adjacent tracks
  static const int F0CHANGE2 = 11; // in Hz, use a looser threshold here
  int maxLabs[] = {lab1, lab2}, cnt, idx, idx1;
  bool keepTrack;
  for (int n = 0; n < 2; n++) {
    for (int t = 0, lab = 1+labInc; lab <= maxLabs[n]; lab += labInc) {
      cnt = idx = 0;
      for ( ; t < nframes; t++ ) {
        if ( labels[t][n] == 0 ) continue;
        if ( labels[t][n] == lab ) { idx = t; cnt++; }
        else break;
      }
      if ( cnt > 2 ) continue;
      keepTrack = false; idx1 = idx - cnt + 1;
      for (int m = 0; m < 2; m++) {
        if ( idx1 > 2 && calc_minDist ( tracks[idx1][n], tracks[idx1-3][m] ) < F0CHANGE2 ) {
          keepTrack = true; break;
        }
        if ( idx < nframes-3 && calc_minDist ( tracks[idx][n], tracks[idx+3][m] ) < F0CHANGE2 ) {
          keepTrack = true; break;
        }
      }
      if ( ! keepTrack ) {
        tracks[idx1][n] = tracks[idx][n] = 0.0;
        labels[idx1][n] = labels[idx][n] = 0;
      }
    }
  }

  // Make labels unique
  for (int t = 0; t < nframes; t++ ) {
    if ( labels[t][1] > 0 ) labels[t][1] += lab1;
  }
}

// Remove the second candidates of the weak source if they're too close to
// other candidates
inline void RuleTracker::remove_closeCands(MatrixF& f0Cands, const int nframes)
{
   for ( int t = 0; t < nframes; t++ ) {
      if ( NUM_F0CANDS > 2 ) {
         if ( fabs ( f0Cands[t][2] - f0Cands[t][0] ) < 6 )
            f0Cands[t][2] = 0.0;
         else if ( fabs ( f0Cands[t][2] - f0Cands[t][1] ) < 6 )
            f0Cands[t][2] = 0.0;
      }
      if ( NUM_F0CANDS > 3 ) {
         if ( fabs ( f0Cands[t][3] - f0Cands[t][0] ) < 6 )
            f0Cands[t][3] = 0.0;
         else if ( fabs ( f0Cands[t][3] - f0Cands[t][1] ) < 6 )
            f0Cands[t][3] = 0.0;
         else if ( fabs ( f0Cands[t][3] - f0Cands[t][2] ) < 6 )
            f0Cands[t][3] = 0.0;
      }
   }
}


// end

