/***************************************************************************
 * An implementation of the fragment generation / multipitch tracking
 * techniques reported in:
 *   Ma, N., Green, P., Barker, J. and Coy, A. (2007) Exploiting correlogram
 *   structure for robust speech recognition with multiple speech sources. 
 *   Speech Communication, 49 (12): 874-891.
 *
 * A demonstration of this technique on simultaneous speech can be found
 * at: http://www.dcs.shef.ac.uk/~ning/pubs/ma_specom07_fig10.html
 *
 * This is a primitive implementation in C++ and requires the "fftw3" 
 * package (http://www.fftw.org) to compute FFT. Any bug report and 
 * suggestions welcome.
 *
 * Ning Ma, University of Sheffield
 * n.ma@dcs.shef.ac.uk, 12 Oct 2007
 ***************************************************************************/

#include <iostream>
#include <fstream>
#include <string>
#include "auxiliary.h"
#include "acgmodel.h"
#include "casagrouping.h"
#include "ruletracker.h"

using namespace std;

int main ( int argc, char** argv )
{
  if (argc != 3) {
    cout << "Usage:" << endl;
    cout << "    " << argv[0] << " INWAVFILE OUTDIR" << endl;
    cout << "eg. " << argv[0] << " s10_bgakzn ./" << endl;
    return 0;
  }
////////////////////////////////////////////////////////////////////////
  AcgParam param;
  param.lowCF = 50.0;    // lowest centre frequency in Hz
  param.highCF = 8000.0; // highest centre frequency in Hz
  param.nchans = 32;     // the number of gammatone filters
  param.winSize = 30;    // ACG window size in millisecond
  param.frmShift = 10;   // frame shift in ms
////////////////////////////////////////////////////////////////////////

  // Extract filename from the full path
  string wavfn(argv[1]);
  string::size_type pos = wavfn.rfind("/");
  if ( pos != string::npos )
    pos++;
  else
    pos = 0;
  string fn = wavfn.substr(pos, wavfn.length() - pos - 4);

  string outdir(argv[2]);
  if ( outdir[outdir.length()-1] != '/' )
    outdir += '/';
  fn = outdir + fn;

  // Test if we can write to the out dir
  string outfn = fn + ".groups";
  ofstream out ( outfn.c_str(), ios::out | ios::binary );
  if ( !out ) {
    cerr << "Unable to write to directory: " << outdir << endl;
    return 0;
  }
  else {
    out.close();
  }
 
  // Read in audio data
  AudioWav wav;
  if ( wav.read(wavfn.c_str()) < 1 )
    return 0;

  // Compute 3D correlogram
  AcgModel acg(param);
  acg.compute(wav.data(), wav.fsHz());
 
  // Generate harmonic fragments
  RuleTracker tracker;
  CASAGrouping grouping(acg, tracker);
  grouping.generate_fragments();

  // Output
  outfn = fn + ".frmseg";    grouping.write_frmSegments(outfn.c_str());  // Segmentations for each frame
  outfn = fn + ".f0cands";   grouping.write_f0Candidates(outfn.c_str()); // F0 candidates
  outfn = fn +  ".f0tracks"; grouping.write_f0Tracks(outfn.c_str());     // Linked pitch tracks
  //outfn = fn + ".f0labs";    grouping.write_f0Labels(outfn.c_str());     // Labels for the pitch tracks
  outfn = fn + ".groups";    grouping.write_fragments(outfn.c_str());    // Fragments corresponding to the pitch tracks
  outfn = fn + ".mask";     grouping.write_mask(outfn.c_str());        // Confidence scores for each time-freq pixel
  return 0;
}

// end
